@@ -1,15 +1,18 @@
 module Imports (
     ByteString(..)
+  , ShortByteString
   , module Control.Applicative
   , module Control.Monad
   , module Data.Bits
+  , module Data.Either
   , module Data.List
   , module Data.Foldable
   , module Data.Int
+  , module Data.Maybe
   , module Data.Monoid
   , module Data.Ord
+  , module Data.String
   , module Data.Word
-  , module Data.Maybe
   , module Numeric
   , GCBuffer
   , withForeignPtr
@@ -20,11 +23,14 @@ import Control.Applicative
 import Control.Monad
 import Data.Bits hiding (Bits)
 import Data.ByteString.Internal (ByteString(..))
+import Data.ByteString.Short (ShortByteString)
+import Data.Either
 import Data.Foldable
 import Data.Int
 import Data.List
 import Data.Maybe
 import Data.Monoid
+import Data.String
 import Data.Ord
 import Data.Word
 import Foreign.ForeignPtr
