@@ -12,6 +12,7 @@ module Network.HTTP2.Arch (
   , module Network.HTTP2.Arch.Status
   , module Network.HTTP2.Arch.Stream
   , module Network.HTTP2.Arch.Types
+  , module Network.HTTP2.Arch.Window
   ) where
 
 import Network.HTTP2.Arch.Config
@@ -27,3 +28,4 @@ import Network.HTTP2.Arch.Sender
 import Network.HTTP2.Arch.Stream
 import Network.HTTP2.Arch.Status
 import Network.HTTP2.Arch.Types
+import Network.HTTP2.Arch.Window
