@@ -1,6 +1,7 @@
 {-# LANGUAGE NamedFieldPuns #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 module Network.HTTP2.Arch.Sender (
     frameSender
@@ -10,16 +11,17 @@ module Network.HTTP2.Arch.Sender (
   , runTrailersMaker
   ) where
 
-import Control.Concurrent.STM
-import qualified Control.Exception as E
 import qualified Data.ByteString as BS
 import Data.ByteString.Builder (Builder)
 import qualified Data.ByteString.Builder.Extra as B
-import Foreign.Ptr (plusPtr)
+import Data.IORef (readIORef, writeIORef)
+import Foreign.Ptr (plusPtr, minusPtr)
 import Network.ByteOrder
+import qualified UnliftIO.Exception as E
+import UnliftIO.STM
 
 import Imports
-import Network.HPACK (setLimitForEncoding, toHeaderTable)
+import Network.HPACK (setLimitForEncoding, toHeaderTable, TokenHeaderList)
 import Network.HTTP2.Arch.Config
 import Network.HTTP2.Arch.Context
 import Network.HTTP2.Arch.EncodeFrame
@@ -29,6 +31,7 @@ import Network.HTTP2.Arch.Manager hiding (start)
 import Network.HTTP2.Arch.Queue
 import Network.HTTP2.Arch.Stream
 import Network.HTTP2.Arch.Types
+import Network.HTTP2.Arch.Window
 import Network.HTTP2.Frame
 
 ----------------------------------------------------------------
@@ -39,89 +42,103 @@ data Leftover = LZero
 
 ----------------------------------------------------------------
 
-{-# INLINE getStreamWindowSize #-}
-getStreamWindowSize :: Stream -> IO WindowSize
-getStreamWindowSize Stream{streamWindow} = readTVarIO streamWindow
-
-{-# INLINE waitStreamWindowSize #-}
-waitStreamWindowSize :: Stream -> IO ()
-waitStreamWindowSize Stream{streamWindow} = atomically $ do
-    w <- readTVar streamWindow
-    check (w > 0)
-
 {-# INLINE waitStreaming #-}
 waitStreaming :: TBQueue a -> IO ()
 waitStreaming tbq = atomically $ do
     isEmpty <- isEmptyTBQueue tbq
-    check (not isEmpty)
+    checkSTM (not isEmpty)
 
 data Switch = C Control
             | O (Output Stream)
             | Flush
 
+wrapException :: E.SomeException -> IO ()
+wrapException se
+  | Just (e :: HTTP2Error) <- E.fromException se = E.throwIO e
+  | otherwise = E.throwIO $ BadThingHappen se
+
 frameSender :: Context -> Config -> Manager -> IO ()
-frameSender ctx@Context{outputQ,controlQ,connectionWindow,encodeDynamicTable}
+frameSender ctx@Context{outputQ,controlQ,encodeDynamicTable,outputBufferLimit}
             Config{..}
-            mgr = loop 0 `E.catch` ignore
+            mgr = loop 0 `E.catch` wrapException
   where
+    ----------------------------------------------------------------
+    loop :: Offset -> IO ()
+    loop off = do
+        x <- atomically $ dequeue off
+        case x of
+            C ctl -> flushN off >> control ctl >> loop 0
+            O out -> outputOrEnqueueAgain out off >>= flushIfNecessary >>= loop
+            Flush -> flushN off >> loop 0
+
+    -- Flush the connection buffer to the socket, where the first 'n' bytes of
+    -- the buffer are filled.
+    flushN :: Offset -> IO ()
+    flushN 0 = return ()
+    flushN n = bufferIO confWriteBuffer n confSendAll
+
+    flushIfNecessary :: Offset -> IO Offset
+    flushIfNecessary off = do
+        buflim <- readIORef outputBufferLimit
+        if off <= buflim - 512 then
+            return off
+          else do
+            flushN off
+            return 0
+
+    dequeue :: Offset -> STM Switch
     dequeue off = do
         isEmpty <- isEmptyTQueue controlQ
         if isEmpty then do
-            w <- readTVar connectionWindow
-            check (w > 0)
+            waitConnectionWindowSize ctx
             emp <- isEmptyTQueue outputQ
             if emp then
-                if off /= 0 then return Flush else retry
+                if off /= 0 then return Flush else retrySTM
               else
                 O <$> readTQueue outputQ
           else
             C <$> readTQueue controlQ
 
-    hardLimit = confBufferSize - 512
+    ----------------------------------------------------------------
+    copyAll []     buf = return buf
+    copyAll (x:xs) buf = copy buf x >>= copyAll xs
 
-    loop off = do
-        x <- atomically $ dequeue off
-        case x of
-            C ctl -> do
-                when (off /= 0) $ flushN off
-                off' <- control ctl off
-                when (off' >= 0) $ loop off'
-            O out -> do
-                off' <- outputOrEnqueueAgain out off
-                case off' of
-                    0                    -> loop 0
-                    _ | off' > hardLimit -> flushN off' >> loop 0
-                      | otherwise        -> loop off'
-            Flush -> flushN off >> loop 0
-
-    control CFinish         _ = return (-1)
-    control (CGoaway frame) _ = confSendAll frame >> return (-1)
-    control (CFrame frame)  _ = confSendAll frame >> return 0
-    control (CSettings frame alist) _ = do
-        confSendAll frame
-        setLimit alist
-        return 0
-    control (CSettings0 frame1 frame2 alist) off = do -- off == 0, just in case
-        let buf = confWriteBuffer `plusPtr` off
-            off' = off + BS.length frame1 + BS.length frame2
-        buf' <- copy buf frame1
-        void $ copy buf' frame2
-        setLimit alist
-        return off'
-
-    {-# INLINE setLimit #-}
-    setLimit alist = case lookup SettingsHeaderTableSize alist of
-        Nothing  -> return ()
-        Just siz -> setLimitForEncoding siz encodeDynamicTable
+    -- called with off == 0
+    control :: Control -> IO ()
+    control (CFinish     e) = E.throwIO e
+    control (CFrames ms xs) = do
+        buf <- copyAll xs confWriteBuffer
+        let off = buf `minusPtr` confWriteBuffer
+        flushN off
+        case ms of
+          Nothing    -> return ()
+          Just peerAlist -> do
+              -- Peer SETTINGS_INITIAL_WINDOW_SIZE
+              updatePeerSettings ctx peerAlist
+              -- Peer SETTINGS_MAX_FRAME_SIZE
+              case lookup SettingsMaxFrameSize peerAlist of
+                Nothing -> return ()
+                Just payloadLen -> do
+                    let dlim = payloadLen + frameHeaderLength
+                        buflim | confBufferSize >= dlim = dlim
+                               | otherwise              = confBufferSize
+                    writeIORef outputBufferLimit buflim
+              -- Peer SETTINGS_HEADER_TABLE_SIZE
+              case lookup SettingsHeaderTableSize peerAlist of
+                Nothing  -> return ()
+                Just siz -> setLimitForEncoding siz encodeDynamicTable
 
+    ----------------------------------------------------------------
+    output :: Output Stream -> Offset -> WindowSize -> IO Offset
     output out@(Output strm OutObj{} (ONext curr tlrmkr) _ sentinel) off0 lim = do
         -- Data frame payload
+        buflim <- readIORef outputBufferLimit
         let payloadOff = off0 + frameHeaderLength
             datBuf     = confWriteBuffer `plusPtr` payloadOff
-            datBufSiz  = confBufferSize - payloadOff
-        Next datPayloadLen mnext <- curr datBuf datBufSiz lim -- checkme
+            datBufSiz  = buflim - payloadOff
+        Next datPayloadLen reqflush mnext <- curr datBuf datBufSiz lim -- checkme
         NextTrailersMaker tlrmkr' <- runTrailersMaker tlrmkr datBuf datPayloadLen
-        fillDataHeaderEnqueueNext strm off0 datPayloadLen mnext tlrmkr' sentinel out
+        fillDataHeaderEnqueueNext strm off0 datPayloadLen mnext tlrmkr' sentinel out reqflush
 
     output out@(Output strm (OutObj hdr body tlrmkr) OObj mtbq _) off0 lim = do
         -- Header frame and Continuation frame
@@ -130,8 +147,8 @@ frameSender ctx@Context{outputQ,controlQ,connectionWindow,encodeDynamicTable}
                 OutBodyNone -> True
                 _           -> False
         (ths,_) <- toHeaderTable $ fixHeaders hdr
-        kvlen <- headerContinue sid ths endOfStream off0
-        off <- sendHeadersIfNecessary $ off0 + frameHeaderLength + kvlen
+        off' <- headerContinue sid ths endOfStream off0
+        off <- flushIfNecessary off'
         case body of
             OutBodyNone -> do
                 -- halfClosedLocal calls closed which removes
@@ -162,12 +179,13 @@ frameSender ctx@Context{outputQ,controlQ,connectionWindow,encodeDynamicTable}
         -- Frame id should be associated stream id from the client.
         let sid = streamNumber strm
         len <- pushPromise pid sid ths off0
-        off <- sendHeadersIfNecessary $ off0 + frameHeaderLength + len
+        off <- flushIfNecessary $ off0 + frameHeaderLength + len
         output out{outputType=OObj} off lim
 
     output _ _ _ = undefined -- never reach
 
-    outputOrEnqueueAgain :: Output Stream -> Int -> IO Int
+    ----------------------------------------------------------------
+    outputOrEnqueueAgain :: Output Stream -> Offset -> IO Offset
     outputOrEnqueueAgain out@(Output strm _ otyp _ _) off = E.handle resetStream $ do
         state <- readStreamState strm
         if isHalfClosedLocal state then
@@ -195,61 +213,64 @@ frameSender ctx@Context{outputQ,controlQ,connectionWindow,encodeDynamicTable}
                 forkAndEnqueueWhenReady (waitStreamWindowSize strm) outputQ out mgr
                 return off
               else do
-                cws <- readTVarIO connectionWindow -- not 0
+                cws <- getConnectionWindowSize ctx -- not 0
                 let lim = min cws sws
                 output out off lim
         resetStream e = do
             closed ctx strm (ResetByMe e)
             let rst = resetFrame InternalError $ streamNumber strm
-            enqueueControl controlQ $ CFrame rst
+            enqueueControl controlQ $ CFrames Nothing [rst]
             return off
 
-    {-# INLINE flushN #-}
-    -- Flush the connection buffer to the socket, where the first 'n' bytes of
-    -- the buffer are filled.
-    flushN :: Int -> IO ()
-    flushN n = bufferIO confWriteBuffer n confSendAll
+    ----------------------------------------------------------------
+    headerContinue :: StreamId -> TokenHeaderList -> Bool -> Offset -> IO Offset
+    headerContinue sid ths0 endOfStream off0 = do
+        buflim <- readIORef outputBufferLimit
+        let offkv = off0 + frameHeaderLength
+            bufkv = confWriteBuffer `plusPtr` offkv
+            limkv = buflim - offkv
+        (ths,kvlen) <- hpackEncodeHeader ctx bufkv limkv ths0
+        if kvlen == 0 then
+            continue off0 ths FrameHeaders
+          else do
+            let flag = getFlag ths
+                buf = confWriteBuffer `plusPtr` off0
+                off = offkv + kvlen
+            fillFrameHeader FrameHeaders kvlen sid flag buf
+            continue off ths FrameContinuation
+      where
+        eos = if endOfStream then setEndStream else id
+        getFlag [] = eos $ setEndHeader defaultFlags
+        getFlag _  = eos $ defaultFlags
 
-    headerContinue sid ths endOfStream off = do
-        let offkv = off + frameHeaderLength
-        let bufkv = confWriteBuffer `plusPtr` offkv
-            limkv = confBufferSize - offkv
-        (hs,kvlen) <- hpackEncodeHeader ctx bufkv limkv ths
-        let flag0 = case hs of
-                [] -> setEndHeader defaultFlags
-                _  -> defaultFlags
-            flag = if endOfStream then setEndStream flag0 else flag0
-        let buf = confWriteBuffer `plusPtr` off
-        fillFrameHeader FrameHeaders kvlen sid flag buf
-        continue sid kvlen hs
+        continue :: Offset -> TokenHeaderList -> FrameType -> IO Offset
+        continue off [] _ = return off
+        continue off ths ft = do
+            flushN off
+            -- Now off is 0
+            buflim <- readIORef outputBufferLimit
+            let bufHeaderPayload = confWriteBuffer `plusPtr` frameHeaderLength
 
-    bufHeaderPayload = confWriteBuffer `plusPtr` frameHeaderLength
-    headerPayloadLim = confBufferSize - frameHeaderLength
+                headerPayloadLim = buflim - frameHeaderLength
+            (ths', kvlen') <- hpackEncodeHeaderLoop ctx bufHeaderPayload headerPayloadLim ths
+            when (ths == ths') $ E.throwIO $ ConnectionErrorIsSent CompressionError sid "cannot compress the header"
+            let flag = getFlag ths'
+                off' = frameHeaderLength + kvlen'
+            fillFrameHeader ft kvlen' sid flag confWriteBuffer
+            continue off' ths' FrameContinuation
 
-    continue _   kvlen [] = return kvlen
-    continue sid kvlen ths = do
-        flushN $ kvlen + frameHeaderLength
-        -- Now off is 0
-        (ths', kvlen') <- hpackEncodeHeaderLoop ctx bufHeaderPayload headerPayloadLim ths
-        when (ths == ths') $ E.throwIO $ ConnectionError CompressionError "cannot compress the header"
-        let flag = case ths' of
-                [] -> setEndHeader defaultFlags
-                _  -> defaultFlags
-        fillFrameHeader FrameContinuation kvlen' sid flag confWriteBuffer
-        continue sid kvlen' ths'
-
-    {-# INLINE sendHeadersIfNecessary #-}
-    -- Send headers if there is not room for a 1-byte data frame, and return
-    -- the offset of the next frame's first header byte.
-    sendHeadersIfNecessary off
-      -- True if the connection buffer has room for a 1-byte data frame.
-      | off + frameHeaderLength < confBufferSize = return off
-      | otherwise = do
-          flushN off
-          return 0
-
-    fillDataHeaderEnqueueNext strm@Stream{streamWindow,streamNumber}
-                   off datPayloadLen Nothing tlrmkr tell _ = do
+    ----------------------------------------------------------------
+    fillDataHeaderEnqueueNext :: Stream
+                              -> Offset
+                              -> Int
+                              -> Maybe DynaNext
+                              -> (Maybe ByteString -> IO NextTrailersMaker)
+                              -> IO ()
+                              -> Output Stream
+                              -> Bool
+                              -> IO Offset
+    fillDataHeaderEnqueueNext strm@Stream{streamNumber}
+                   off datPayloadLen Nothing tlrmkr tell _ reqflush = do
         let buf  = confWriteBuffer `plusPtr` off
             off' = off + frameHeaderLength + datPayloadLen
         (mtrailers, flag) <- do
@@ -262,34 +283,45 @@ frameSender ctx@Context{outputQ,controlQ,connectionWindow,encodeDynamicTable}
         off'' <- handleTrailers mtrailers off'
         void tell
         when (isServer ctx) $ halfClosedLocal ctx strm Finished
-        atomically $ modifyTVar' connectionWindow (subtract datPayloadLen)
-        atomically $ modifyTVar' streamWindow (subtract datPayloadLen)
-        return off''
+        decreaseWindowSize ctx strm datPayloadLen
+        if reqflush then do
+            flushN off''
+            return 0
+          else
+            return off''
       where
         handleTrailers Nothing off0 = return off0
         handleTrailers (Just trailers) off0 = do
             (ths,_) <- toHeaderTable trailers
-            kvlen <- headerContinue streamNumber ths True off0
-            sendHeadersIfNecessary $ off0 + frameHeaderLength + kvlen
+            headerContinue streamNumber ths True off0
 
     fillDataHeaderEnqueueNext _
-                   off 0 (Just next) tlrmkr _ out = do
+                   off 0 (Just next) tlrmkr _ out reqflush = do
         let out' = out { outputType = ONext next tlrmkr }
         enqueueOutput outputQ out'
-        return off
+        if reqflush then do
+            flushN off
+            return 0
+          else
+            return off
 
-    fillDataHeaderEnqueueNext Stream{streamWindow,streamNumber}
-                   off datPayloadLen (Just next) tlrmkr _ out = do
+    fillDataHeaderEnqueueNext strm@Stream{streamNumber}
+                   off datPayloadLen (Just next) tlrmkr _ out reqflush = do
         let buf  = confWriteBuffer `plusPtr` off
             off' = off + frameHeaderLength + datPayloadLen
             flag  = defaultFlags
         fillFrameHeader FrameData datPayloadLen streamNumber flag buf
-        atomically $ modifyTVar' connectionWindow (subtract datPayloadLen)
-        atomically $ modifyTVar' streamWindow (subtract datPayloadLen)
+        decreaseWindowSize ctx strm datPayloadLen
         let out' = out { outputType = ONext next tlrmkr }
         enqueueOutput outputQ out'
-        return off'
+        if reqflush then do
+            flushN off'
+            return 0
+          else
+            return off'
 
+    ----------------------------------------------------------------
+    pushPromise :: StreamId -> StreamId -> TokenHeaderList -> Offset -> IO Int
     pushPromise pid sid ths off = do
         let offsid = off + frameHeaderLength -- checkme
             bufsid = confWriteBuffer `plusPtr` offsid
@@ -304,14 +336,16 @@ frameSender ctx@Context{outputQ,controlQ,connectionWindow,encodeDynamicTable}
         fillFrameHeader FramePushPromise len pid flag buf
         return len
 
+    ----------------------------------------------------------------
     {-# INLINE fillFrameHeader #-}
+    fillFrameHeader :: FrameType -> Int -> StreamId -> FrameFlags -> Buffer -> IO ()
     fillFrameHeader ftyp len sid flag buf = encodeFrameHeaderBuf ftyp hinfo buf
       where
-        hinfo = FrameHeader len flag sid
-
-    {-# INLINE ignore #-}
-    ignore :: E.SomeException -> IO ()
-    ignore _ = return ()
+        hinfo = FrameHeader {
+            payloadLength = len
+          , flags         = flag
+          , streamId      = sid
+          }
 
 -- | Running trailers-maker.
 --
@@ -337,8 +371,8 @@ fillFileBodyGetNext pread start bytecount refresh buf siz lim = do
 fillStreamBodyGetNext :: IO (Maybe StreamingChunk) -> DynaNext
 fillStreamBodyGetNext takeQ buf siz lim = do
     let room = min siz lim
-    (leftover, cont, len) <- runStreamBuilder buf room takeQ
-    return $ nextForStream takeQ leftover cont len
+    (cont, len, reqflush, leftover) <- runStreamBuilder buf room takeQ
+    return $ nextForStream cont len reqflush leftover takeQ
 
 ----------------------------------------------------------------
 
@@ -365,39 +399,42 @@ fillBufBuilder leftover buf0 siz0 lim = do
 
 nextForBuilder :: BytesFilled -> B.Next -> Next
 nextForBuilder len B.Done
-    = Next len Nothing
+    = Next len True Nothing -- let's flush
 nextForBuilder len (B.More _ writer)
-    = Next len $ Just (fillBufBuilder (LOne writer))
+    = Next len False $ Just (fillBufBuilder (LOne writer))
 nextForBuilder len (B.Chunk bs writer)
-    = Next len $ Just (fillBufBuilder (LTwo bs writer))
+    = Next len False $ Just (fillBufBuilder (LTwo bs writer))
 
 ----------------------------------------------------------------
 
 runStreamBuilder :: Buffer -> BufferSize -> IO (Maybe StreamingChunk)
-                 -> IO (Leftover, Bool, BytesFilled)
+                 -> IO (Bool  -- continue
+                       ,BytesFilled
+                       ,Bool  -- require flusing
+                       ,Leftover)
 runStreamBuilder buf0 room0 takeQ = loop buf0 room0 0
   where
     loop buf room total = do
         mbuilder <- takeQ
         case mbuilder of
-            Nothing      -> return (LZero, True, total)
+            Nothing      -> return (True, total, False, LZero)
             Just (StreamingBuilder builder) -> do
                 (len, signal) <- B.runBuilder builder buf room
                 let total' = total + len
                 case signal of
                     B.Done -> loop (buf `plusPtr` len) (room - len) total'
-                    B.More  _ writer  -> return (LOne writer, True, total')
-                    B.Chunk bs writer -> return (LTwo bs writer, True, total')
-            Just StreamingFlush       -> return (LZero, True, total)
-            Just StreamingFinished    -> return (LZero, False, total)
+                    B.More  _ writer  -> return (True,  total', False, LOne writer)
+                    B.Chunk bs writer -> return (True,  total', False, LTwo bs writer)
+            Just StreamingFlush       -> return (True,  total,  True,  LZero)
+            Just StreamingFinished    -> return (False, total,  True,  LZero)
 
 fillBufStream :: Leftover -> IO (Maybe StreamingChunk) -> DynaNext
 fillBufStream leftover0 takeQ buf0 siz0 lim0 = do
     let room0 = min siz0 lim0
     case leftover0 of
         LZero -> do
-            (leftover, cont, len) <- runStreamBuilder buf0 room0 takeQ
-            getNext leftover cont len
+            (cont, len, reqflush, leftover) <- runStreamBuilder buf0 room0 takeQ
+            getNext cont len reqflush leftover
         LOne writer -> write writer buf0 room0 0
         LTwo bs writer
           | BS.length bs <= room0 -> do
@@ -407,29 +444,34 @@ fillBufStream leftover0 takeQ buf0 siz0 lim0 = do
           | otherwise -> do
               let (bs1,bs2) = BS.splitAt room0 bs
               void $ copy buf0 bs1
-              getNext (LTwo bs2 writer) True room0
+              getNext True room0 False $ LTwo bs2 writer
   where
-    getNext l b r = return $ nextForStream takeQ l b r
+    getNext :: Bool -> BytesFilled -> Bool -> Leftover -> IO Next
+    getNext cont len reqflush l = return $ nextForStream cont len reqflush l takeQ
+
+    write :: (Buffer -> BufferSize -> IO (Int, B.Next))
+          -> Buffer -> BufferSize -> Int
+          -> IO Next
     write writer1 buf room sofar = do
         (len, signal) <- writer1 buf room
         case signal of
             B.Done -> do
-                (leftover, cont, extra) <- runStreamBuilder (buf `plusPtr` len) (room - len) takeQ
+                (cont, extra, reqflush, leftover) <- runStreamBuilder (buf `plusPtr` len) (room - len) takeQ
                 let total = sofar + len + extra
-                getNext leftover cont total
+                getNext cont total reqflush leftover
             B.More  _ writer -> do
                 let total = sofar + len
-                getNext (LOne writer) True total
+                getNext True total False $ LOne writer
             B.Chunk bs writer -> do
                 let total = sofar + len
-                getNext (LTwo bs writer) True total
+                getNext True total False $ LTwo bs writer
 
-nextForStream :: IO (Maybe StreamingChunk)
-              -> Leftover -> Bool -> BytesFilled
+nextForStream :: Bool -> BytesFilled -> Bool
+              -> Leftover -> IO (Maybe StreamingChunk)
               -> Next
-nextForStream _ _ False len = Next len Nothing
-nextForStream takeQ leftOrZero True len =
-    Next len $ Just (fillBufStream leftOrZero takeQ)
+nextForStream False len reqflush _          _     = Next len reqflush Nothing
+nextForStream True  len reqflush leftOrZero takeQ =
+    Next len reqflush $ Just (fillBufStream leftOrZero takeQ)
 
 ----------------------------------------------------------------
 
@@ -442,10 +484,10 @@ fillBufFile pread start bytes refresh buf siz lim = do
     return $ nextForFile len' pread (start + len) (bytes - len) refresh
 
 nextForFile :: BytesFilled -> PositionRead -> FileOffset -> ByteCount -> IO () -> Next
-nextForFile 0   _  _     _     _       = Next 0   Nothing
-nextForFile len _  _     0     _       = Next len Nothing
+nextForFile 0   _  _     _     _       = Next 0   True  Nothing -- let's flush
+nextForFile len _  _     0     _       = Next len False Nothing
 nextForFile len pread start bytes refresh =
-    Next len $ Just (fillBufFile pread start bytes refresh)
+    Next len False $ Just $ fillBufFile pread start bytes refresh
 
 {-# INLINE mini #-}
 mini :: Int -> Int64 -> Int64
